/*
 * Copyright (c) 1995, the EUROPAGATE consortium (see below).
 *
 * The EUROPAGATE consortium members are:
 *
 *    University College Dublin
 *    Danmarks Teknologiske Videnscenter
 *    An Chomhairle Leabharlanna
 *    Consejo Superior de Investigaciones Cientificas
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation, in whole or in part, for any purpose, is hereby granted,
 * provided that:
 *
 * 1. This copyright and permission notice appear in all copies of the
 * software and its documentation. Notices of copyright or attribution
 * which appear at the beginning of any file must remain unchanged.
 *
 * 2. The names of EUROPAGATE or the project partners may not be used to
 * endorse or promote products derived from this software without specific
 * prior written permission.
 *
 * 3. Users of this software (implementors and gateway operators) agree to
 * inform the EUROPAGATE consortium of their use of the software. This
 * information will be used to evaluate the EUROPAGATE project and the
 * software, and to plan further developments. The consortium may use
 * the information in later publications.
 * 
 * 4. Users of this software agree to make their best efforts, when
 * documenting their use of the software, to acknowledge the EUROPAGATE
 * consortium, and the role played by the software in their work.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED, OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL THE EUROPAGATE CONSORTIUM OR ITS MEMBERS BE LIABLE
 * FOR ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF
 * ANY KIND, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND
 * ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */
/* Gateway utility
 * Europagate, 1995
 *
 * $Log: gw-db.c,v $
 * Revision 1.5  1996/01/24 17:00:12  adam
 * Bug fix: missing argument to gw_log-call.
 *
 * Revision 1.4  1995/12/20  16:28:07  adam
 * Extra parameter block to gw_db_open. If block is 0 gw_db_open
 * returns NULL if lock couldn't be satisfied.
 * Minor changes in iso2709.c.
 *
 * Revision 1.3  1995/05/16  09:40:53  adam
 * LICENSE.
 *
 * Revision 1.2  1995/05/01  12:43:57  adam
 * lgets function moved from kernel to util.
 *
 * Revision 1.1  1995/03/27  08:25:01  adam
 * New module gip: Gateway IPc module.
 * New module gw-db: Gateway hash-db module (user information table).
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <gw-log.h>
#include <gw-db.h>

#define FILE_HEAD_LEN 128
#define DB_HASH 1997

struct gw_db {
    struct file_head {     /* file header info */
        char magic[8];     /* start info */
	int no_of_entries; /* no of entries in table */
	int sequence_no;   /* number of next to come */
    } head;
    int *hash_array;       /* hash array pointers (DB_HASH in size) */
    int fd;                /* descriptor of file */
    int dirty;             /* any writes so far? */
};

struct db_bucket {         /* information about individual entries */
    int  name_length;      /* length of name */
    int  info_length;      /* length of information */
    int  next;             /* next in hash chain - possibly 0 */
    int  prev;             /* prev in hach chain - possibly 0 */
};

static char *mod = "gwdb";

/*
 * write_head: write header of table
 */
static int write_head (GW_DB db)
{
    char file_head_buf[FILE_HEAD_LEN];
    int r;

    if (lseek (db->fd, 0L, SEEK_SET) == -1)
        return -1;
    memcpy (file_head_buf, &db->head, sizeof(db->head));
    r = write (db->fd, file_head_buf, FILE_HEAD_LEN);
    if (r == -1)
        return -1;
    r = write (db->fd, db->hash_array, DB_HASH * sizeof(*db->hash_array));
    if (r == -1)
        return -1;
    return 0;
}

/*
 * hash: calculate hash value
 */
static unsigned hash (const char *name, int length)
{
    unsigned l = 0;

    while (--length >= 0)
        l = l*65599 + *name++;
    return l % DB_HASH;
}

/*
 * lock_file: lock entire file 
 */
static int lock_file (int fd, int block, int type)
{
    struct flock area;
    area.l_type = type;
    area.l_whence = SEEK_SET;
    area.l_start = 0L;
    area.l_len = 0L;
    return fcntl (fd, block ? F_SETLKW : F_SETLK, &area);
}


/*
 * gw_db_lookup: table lookup
 */
int gw_db_lookup (GW_DB db, const char *name, int name_length,
                  void **buf, size_t *count)
{
    struct db_bucket bucket;
    unsigned l = hash (name, name_length);
    char     *dbuf = NULL;
    int      dsize = 0;
    int      pos;
    int      r;

    for (pos = db->hash_array[l]; pos; pos = bucket.next)
    {
        int blen;

        if (lseek (db->fd, pos, SEEK_SET) == -1)
	{
	    free (dbuf);
	    return -1;
	}
        r = read (db->fd, &bucket, sizeof(bucket));
	if (r == -1)
	{
	    free (dbuf);
	    return -1;
	}
	if (r != sizeof(bucket))
	{
	    free (dbuf);
	    return -2;
	}
	if (bucket.name_length <= 0 || bucket.info_length <= 0 ||
	    bucket.name_length >= 16384 || bucket.info_length >= 262144)
	{
	    free (dbuf);
	    return -3;
	}
	if (bucket.name_length != name_length)
	    continue;
	blen = bucket.name_length + bucket.info_length;
	if (blen >= dsize)
        {
	    dsize = blen + 1024;
	    free (dbuf);
	    if (!(dbuf = malloc (dsize)))
	        return -1;
        }
	r = read (db->fd, dbuf, blen);
	if (r == -1)
	{
	    free (dbuf);
	    return -1;
	}
	else if (r < blen)
	{
	    free (dbuf);
	    return -2;
	}
	if (memcmp (name, dbuf, name_length))
	    continue;

        *count = bucket.info_length;
        *buf = malloc (*count);
	if (!*buf)
	{
	    free (dbuf);
	    return -1;
	}
	memcpy (*buf, dbuf + name_length, *count);
	free (dbuf);
        return 1;
    }
    free (dbuf);
    return 0;
}

/*
 * gw_db_insert: table insertion
 */
int gw_db_insert (GW_DB db, const char *name, int name_length,
                  const void *buf, size_t count)
{
    struct db_bucket n_bucket;
    struct db_bucket bucket;
    off_t    n_pos, r_pos;
    unsigned l = hash (name, name_length);
    int      pos = db->hash_array[l];
    int      r;

    db->dirty = 1;
    (db->head.no_of_entries)++;
    (db->head.sequence_no)++;
    n_bucket.name_length = name_length;
    n_bucket.info_length = count;
    n_pos = lseek (db->fd, 0, SEEK_END);
    if (n_pos == -1)
        return -1;
    n_bucket.next = pos;
    db->hash_array[l] = n_pos;
    n_bucket.prev = 0;

    r = write (db->fd, &n_bucket, sizeof(n_bucket));
    if (r == -1)
        return -1;
    else if (r < sizeof(n_bucket))
        return -2;

    r = write (db->fd, name, name_length);
    if (r == -1)
        return -1;
    else if (r < name_length)
        return -2;

    r = write (db->fd, buf, count);
    if (r == -1)
        return -1;
    else if (r < count)
        return -2;

    r_pos = lseek (db->fd, pos, SEEK_SET);
    if (r_pos == -1)
        return -1;
    r = read (db->fd, &bucket, sizeof(bucket));
    if (r == -1)
        return -1;
    else if (r < sizeof(bucket))
        return -2;
    bucket.prev = n_pos;

    r_pos = lseek (db->fd, pos, SEEK_SET);
    if (r_pos == -1)
        return -1;
    r = write (db->fd, &bucket, sizeof(bucket));
    if (r == -1)
        return -1;
    else if (r < sizeof(bucket))
        return -2;
    return 0;
}

/*
 * gw_db_free: release memory associated with table
 */
static GW_DB gw_db_free (GW_DB db)
{
    free (db->hash_array);
    free (db);
    return NULL;
}

/*
 * gw_db_open: open and lock table for reading (and writing)
 */
GW_DB gw_db_open (const char *fname, int write_flag, int block)
{
    char file_head_buf[FILE_HEAD_LEN];
    GW_DB db;
    int r;

    db = malloc (sizeof(*db));
    if (!db)
    {
        gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "malloc");
        exit (1);
    }
    db->dirty = 0;
    if (!(db->hash_array = malloc (DB_HASH * sizeof(*db->hash_array))))
    {
        gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "malloc");
        exit (1);
    }
    if (write_flag)
    {
        if ((db->fd = open (fname, O_RDWR|O_CREAT, 0666)) == -1)
        {
            gw_log (GW_LOG_FATAL|GW_LOG_ERRNO, mod, "open %s", fname);
            exit (1);
        }
    }
    else
    {
        if ((db->fd = open (fname, O_RDONLY)) == -1)
        {
            gw_log (GW_LOG_WARN|GW_LOG_ERRNO, mod, " open %s", fname);
            return gw_db_free (db);
        }
    }
    if (block)
        lock_file (db->fd, 1, write_flag ? F_WRLCK : F_RDLCK);
    else
    {
        r = lock_file (db->fd, 0, write_flag ? F_WRLCK : F_RDLCK);
        if (r == -1)
        {
            gw_log (GW_LOG_WARN|GW_LOG_ERRNO, mod, "flock %s", fname);
            close (db->fd);
            return gw_db_free (db);
        }
    }
    r = read (db->fd, file_head_buf, FILE_HEAD_LEN);
    if (r == -1)
        return gw_db_free (db);
    if (r < FILE_HEAD_LEN)
    {
        int i;

        if (!write_flag)
	    return gw_db_free (db);
        db->head.no_of_entries = 0;
        db->head.sequence_no = 1;
        for (i=0; i<DB_HASH; i++)
	    db->hash_array[i] = 0;
	if (write_head (db))
	    return gw_db_free (db);
    }
    else
    {
        memcpy (&db->head, file_head_buf, sizeof(db->head));
        r = read (db->fd, db->hash_array, sizeof(*db->hash_array)*DB_HASH);
	if (r < sizeof(*db->hash_array)*DB_HASH)
	    return gw_db_free (db);
    }
    return db;
}

/*
 * gw_db_close: close table and flush
 */
int gw_db_close (GW_DB db)
{
    int r = 0;

    if (db->dirty) 
        r = write_head (db);
    lock_file (db->fd, 1, F_UNLCK);
    if (close (db->fd) == -1)
    {
        gw_db_free (db);
        return -1;
    }
    gw_db_free (db);
    return r;
}

/*
 * gw_db_no_ent: return number of entries in table
 */
int gw_db_no_ent (GW_DB db)
{
    return db->head.no_of_entries;
}

/*
 * gw_db_seq_no: return sequence number
 */
int gw_db_seq_no (GW_DB db)
{
    return db->head.sequence_no;
}

